;;;; Lambda Calculus abstractions.
;
(define true  (lambda (x y) 
                x))
(define false  (lambda (x y) 
                 y))
(define lif  (lambda (b t f) 
                (b t f)))
(define lnot  (lambda (b) 
               (b false true )))
(define land  (lambda (x y) 
               (lif x y x)))
(define lor  (lambda (x y) 
               (lif x x y)))
(define equalp  (lambda (m n) 
               (land (zerop (sub m n))
                     (zerop (sub n m))))) 
(define gtp  (lambda (m n) 
               (lnot (zerop (sub m n)))))
(define ltp  (lambda (m n) 
               (lnot (zerop (sub n m)))))
(define gep  (lambda (m n) 
               (zerop (sub n m))))
(define lcons  (lambda (x y) 
                 (lambda (f) 
                   (f x y))))
(define lcar  (lambda (l) 
                (l true)))
(define lcdr  (lambda (l) 
                (l false)))
(define llength  (lambda (l) 
                   (lif (nullp l) 
                        zero
                        (add one (llength (lcdr l))))))
(define zero  (lambda (f) 
                (lambda (x) 
                  x)))
(define one  (lambda (f) 
                (lambda (x) 
                  (f x))))
(define two  (lambda (f) 
                (lambda (x) 
                  (f (f x)))))
(define three  (lambda (f) 
                 (lambda (x) 
                   (f (f (f x))))))
(define compose  (lambda (f g)
                   (lambda (x) 
                     (f (g x)))))
(define add  (lambda (m n)
               (lambda (f) 
                 (compose (m f) (n f)))))
(define succ  (lambda (n) 
                (lambda (f) 
                  (compose  f (n f)))))
(define mult  (lambda (m n)
                (compose m n)))
(define zeropair  (lcons zero zero))
(define pred  (lambda (n) 
                (lcdr ((n (lambda (x) 
                             (lcons (add (lcar x) one)
                                    (lcar x))))
                       zeropair))))
(define sub  (lambda (m n)
               ((n pred) m)))
(define zerop  (lambda (n)
                 ((n (lambda(x) false)) true)))


(define nil (lambda (l)
              true))

(define nullp (lambda (l)
                (l (lambda (a d)
                     false))))

(define curry
  (lambda(f)
    (lambda(x)
      (lambda(y)
        (f x y)))))

(define map
  (lambda(f l)
    (lif (nullp l)
        nil
        (lcons (f (lcar l)) (map f (lcdr l))))))

(define filter
  (lambda(p l)
    (lif (nullp l)
        nil
        (lif (p (lcar l))
             (lcons (lcar l) (filter p (lcdr l)))
             (filter p (lcdr l))))))

(define mapc (curry map))

(define succ* (mapc succ))


(define addelt
  (lambda(x s)
    (lif (memberp x s)
        s
        (lcons x s))))

(define union
  (lambda(s1 s2)
    (lif (nullp s1) 
        s2
        (lif (memberp (lcar s1) s2) 
            (union (lcdr s1) s2)
            (lcons (lcar s1) (union (lcdr s1) s2))))))

(define memberp
  (lambda(x s)
    (lif (nullp s)
         false
         (lif (equalp x (lcar s))
              true
              (memberp x (lcdr s))))))


(define insert 
  (lambda(x l)
    (lif (nullp l)
         (lcons x nil)
         (lif (ltp x (lcar l))
             (lcons x l)
             (lcons (lcar l) (insert x (lcdr l)))))))

(define insertion-sort
  (lambda(l)
    (lif (nullp l)
        nil
        (insert (lcar l) (insertion-sort (lcdr l))))))

(define sum 
  (lambda(l)
    (lif (nullp l)
        zero
        (add (lcar l) (sum (lcdr l))))))

;;(define nonterminate 
;;          ((lambda(x) (x x)) (lambda(x) (x x))))
;
;;;; the following functions provide support from the interpreter
;;;;
;
(define ndisp! (lambda (n)
                (print ((n incr) vmzero))))

(define bdisp! (lambda (b)
                (print (b vmtrue vmfalse))))
;
(define ldisp! (lambda (l)
                (lif (nullp l) 
                    nil
                    ((lambda()
                       (ndisp! (lcar l))
                       (ldisp! (lcdr l)))))))
;;;;;;;; applications
;;;;;;;;
;;;;;
(define four (succ three))
(define five (succ four))
(define six (mult two three))
(define seven (add three four))
(define eight (add four four))
(define nine (add four five))
(define ten (add five five))
;;;;;
;;;;;; lif uses lazy evaluation !!
;;;;;;
(define if1
   (lambda(m n)
     (lif (equalp m n)
          vmtrue
          ((lambda(x) (x x)) (lambda(x) (x x))))))

;;;;;
;;;;;
;;;;;                 --> 0
;;;;;                     1
;;;;;                     2
;;;;;                     3
;;;;;                     4
;;;
(define l1
   (lcons four (lcons three (lcons two (lcons one nil)))))
;;;;;
(ndisp! (llength l1))
;;;;;                 --> 4
(ldisp! l1)
;;;;;                 --> 4
;;;;;                     3
;;;;;                     2
;;;;;                     1
;;;;;
;(insertion-sort l1)
(ldisp! (insertion-sort l1))
;;;;;                 --> 1
;;;;;                     2
;;;;;                     3
;;;;;                     4
;
(ndisp! (sum l1))
;;;;;                 --> 10
;
(succ* l1)
(ldisp! (succ* l1))
;;;;;                 --> 5
;;;;;                     4
;;;;;                     3
;;;;;                     2
;
(filter (lambda(x) (gtp x two)) l1)
(ldisp! (filter (lambda(x) (gtp x two)) l1))
;;;;;                 --> 4
;;;;;                     3          
;
(define while
  (lambda(c body)
    (define loop
      (lambda()
        (lif c
             ((lambda()
                (body)
                (loop)))
             false)))
    (loop)))

(define while-test
  (lambda(n)
    (while (gtp n zero)
      (lambda()
        (define n (pred n))
        (ndisp! n)))))

(while-test ten)

;;;;
;;;;
;;;; Objekt-Orientierung in ARS
;;;;
;;;;

(define make-account 
  (lambda(balance)
    (define get-balance
      (lambda()
        balance))

    (define deposit 
      (lambda(amount)
        (define balance (add balance amount))
        balance))

    (define withdraw 
      (lambda(amount)
        (lif (gep balance amount)
             ((lambda()
                (define balance (sub balance amount))
                balance))
             false)))

    (define print-account
      (lambda()
        (ndisp! balance)))

    (define self 
      (lambda(msg)
        (lif (equalp msg one) 
             get-balance
             (lif (equalp msg two) 
                  deposit
                  (lif (equalp msg three) 
                       withdraw
                       (lif (equalp msg four) 
                            print-account
                            false)))))) 

    self))

(define konto
  (lambda()
    (define k1 (make-account ten))
    (define k2 (make-account five))
    (define msg-deposit two)
    (define msg-withdraw three)
    (define msg-print four)
    
    ((k1 msg-print))   
    ((k1 msg-deposit) ten)
    ((k1 msg-print))   
    ((k1 msg-withdraw) four)   
    ((k1 msg-print))   

    ((k2 msg-print))   
    ((k2 msg-deposit) nine)   
    ((k2 msg-print))   
    ((k2 msg-withdraw) seven) 
    ((k2 msg-print))))
 
 (konto)
 ;(quit)
